@@ -1,4 +1,5 @@
 import logging
+import typing
 from collections import OrderedDict
 
 from lxml import etree
@@ -7,6 +8,8 @@ from zeep import exceptions, ns
 from zeep.loader import load_external
 from zeep.settings import Settings
 from zeep.xsd import const
+from zeep.xsd import elements as xsd_elements
+from zeep.xsd import types as xsd_types
 from zeep.xsd.elements import builtins as xsd_builtins_elements
 from zeep.xsd.types import builtins as xsd_builtins_types
 from zeep.xsd.visitor import SchemaVisitor
@@ -14,7 +17,7 @@ from zeep.xsd.visitor import SchemaVisitor
 logger = logging.getLogger(__name__)
 
 
-class Schema(object):
+class Schema:
     """A schema is a collection of schema documents."""
 
     def __init__(self, node=None, transport=None, location=None, settings=None):
@@ -44,25 +47,24 @@ class Schema(object):
     def __repr__(self):
         main_doc = self.root_document
         if main_doc:
-            return '<Schema(location=%r, tns=%r)>' % (
-                main_doc._location, main_doc._target_namespace)
-        return '<Schema()>'
+            return "<Schema(location=%r, tns=%r)>" % (
+                main_doc._location,
+                main_doc._target_namespace,
+            )
+        return "<Schema()>"
 
     @property
     def prefix_map(self):
         retval = {}
         retval.update(self._prefix_map_custom)
-        retval.update({
-            k: v for k, v in self._prefix_map_auto.items()
-            if v not in retval.values()
-        })
+        retval.update(
+            {k: v for k, v in self._prefix_map_auto.items() if v not in retval.values()}
+        )
         return retval
 
     @property
     def root_document(self):
-        return next(
-            (doc for doc in self.documents if not doc._is_internal),
-            None)
+        return next((doc for doc in self.documents if not doc._is_internal), None)
 
     @property
     def is_empty(self):
@@ -101,14 +103,9 @@ class Schema(object):
                     yield type_
                     seen.add(type_.qname)
 
-    def add_documents(self, schema_nodes, location):
-        """
-
-        :type schema_nodes: List[lxml.etree._Element]
-        :type location: str
-        :type target_namespace: Optional[str]
-
-        """
+    def add_documents(
+        self, schema_nodes: typing.List[etree._Element], location: str
+    ) -> None:
         resolve_queue = []
         for node in schema_nodes:
             document = self.create_new_document(node, location)
@@ -119,21 +116,15 @@ class Schema(object):
 
         self._prefix_map_auto = self._create_prefix_map()
 
-    def add_document_by_url(self, url):
-        schema_node = load_external(
-            url, self._transport, settings=self.settings)
-
+    def add_document_by_url(self, url: str) -> None:
+        schema_node = load_external(url, self._transport, settings=self.settings)
         document = self.create_new_document(schema_node, url=url)
         document.resolve()
 
-    def get_element(self, qname):
-        """Return a global xsd.Element object with the given qname
-
-        :rtype: zeep.xsd.Group
-
-        """
+    def get_element(self, qname) -> xsd_elements.Element:
+        """Return a global xsd.Element object with the given qname"""
         qname = self._create_qname(qname)
-        return self._get_instance(qname, 'get_element', 'element')
+        return self._get_instance(qname, "get_element", "element")
 
     def get_type(self, qname, fail_silently=False):
         """Return a global xsd.Type object with the given qname
@@ -144,36 +135,24 @@ class Schema(object):
         """
         qname = self._create_qname(qname)
         try:
-            return self._get_instance(qname, 'get_type', 'type')
+            return self._get_instance(qname, "get_type", "type")
         except exceptions.NamespaceError as exc:
             if fail_silently:
                 logger.debug(str(exc))
             else:
                 raise
 
-    def get_group(self, qname):
-        """Return a global xsd.Group object with the given qname.
+    def get_group(self, qname) -> xsd_elements.Group:
+        """Return a global xsd.Group object with the given qname."""
+        return self._get_instance(qname, "get_group", "group")
 
-        :rtype: zeep.xsd.Group
+    def get_attribute(self, qname) -> xsd_elements.Attribute:
+        """Return a global xsd.attribute object with the given qname"""
+        return self._get_instance(qname, "get_attribute", "attribute")
 
-        """
-        return self._get_instance(qname, 'get_group', 'group')
-
-    def get_attribute(self, qname):
-        """Return a global xsd.attributeGroup object with the given qname
-
-        :rtype: zeep.xsd.Attribute
-
-        """
-        return self._get_instance(qname, 'get_attribute', 'attribute')
-
-    def get_attribute_group(self, qname):
-        """Return a global xsd.attributeGroup object with the given qname
-
-        :rtype: zeep.xsd.AttributeGroup
-
-        """
-        return self._get_instance(qname, 'get_attribute_group', 'attributeGroup')
+    def get_attribute_group(self, qname) -> xsd_elements.AttributeGroup:
+        """Return a global xsd.attributeGroup object with the given qname"""
+        return self._get_instance(qname, "get_attribute_group", "attributeGroup")
 
     def set_ns_prefix(self, prefix, namespace):
         self._prefix_map_custom[prefix] = namespace
@@ -195,8 +174,8 @@ class Schema(object):
             if namespace == other_namespace:
                 return prefix
 
-        if namespace == 'http://schemas.xmlsoap.org/soap/envelope/':
-            return 'soap-env'
+        if namespace == "http://schemas.xmlsoap.org/soap/envelope/":
+            return "soap-env"
         return namespace
 
     def create_new_document(self, node, url, base_url=None, target_namespace=None):
@@ -205,11 +184,12 @@ class Schema(object):
         :rtype: zeep.xsd.schema.SchemaDocument
 
         """
-        namespace = node.get('targetNamespace') if node is not None else None
+        namespace = node.get("targetNamespace") if node is not None else None
         if not namespace:
             namespace = target_namespace
         if base_url is None:
             base_url = url
+
         schema = SchemaDocument(namespace, url, base_url)
         self.documents.add(schema)
         schema.load(self, node)
@@ -244,7 +224,7 @@ class Schema(object):
         """Return an object from one of the SchemaDocument's"""
         qname = self._create_qname(qname)
         try:
-            last_exception = None
+            last_exception = None  # type: typing.Optional[BaseException]
             for schema in self._get_schema_documents(qname.namespace):
                 method = getattr(schema, method_name)
                 try:
@@ -252,13 +232,17 @@ class Schema(object):
                 except exceptions.LookupError as exc:
                     last_exception = exc
                     continue
-            raise last_exception
+            if last_exception is not None:
+                raise last_exception
 
         except exceptions.NamespaceError:
-            raise exceptions.NamespaceError((
-                "Unable to resolve %s %s. " +
-                "No schema available for the namespace %r."
-            ) % (name, qname.text, qname.namespace))
+            raise exceptions.NamespaceError(
+                (
+                    "Unable to resolve %s %s. "
+                    + "No schema available for the namespace %r."
+                )
+                % (name, qname.text, qname.namespace)
+            )
 
     def _create_qname(self, name):
         """Create an `lxml.etree.QName()` object for the given qname string.
@@ -271,28 +255,25 @@ class Schema(object):
         if isinstance(name, etree.QName):
             return name
 
-        if not name.startswith('{') and ':' in name and self._prefix_map_auto:
-            prefix, localname = name.split(':', 1)
+        if not name.startswith("{") and ":" in name and self._prefix_map_auto:
+            prefix, localname = name.split(":", 1)
             if prefix in self._prefix_map_custom:
                 return etree.QName(self._prefix_map_custom[prefix], localname)
             elif prefix in self._prefix_map_auto:
                 return etree.QName(self._prefix_map_auto[prefix], localname)
             else:
-                raise ValueError(
-                    "No namespace defined for the prefix %r" % prefix)
+                raise ValueError("No namespace defined for the prefix %r" % prefix)
         else:
             return etree.QName(name)
 
     def _create_prefix_map(self):
-        prefix_map = {
-            'xsd': 'http://www.w3.org/2001/XMLSchema',
-        }
+        prefix_map = {"xsd": "http://www.w3.org/2001/XMLSchema"}
         i = 0
         for namespace in self.documents.get_all_namespaces():
             if namespace is None or namespace in prefix_map.values():
                 continue
 
-            prefix_map['ns%d' % i] = namespace
+            prefix_map["ns%d" % i] = namespace
             i += 1
         return prefix_map
 
@@ -312,7 +293,7 @@ class Schema(object):
         return self.documents.get_by_namespace(namespace, fail_silently)
 
 
-class _SchemaContainer(object):
+class _SchemaContainer:
     """Container instances to store multiple SchemaDocument objects per
     namespace.
 
@@ -325,40 +306,39 @@ class _SchemaContainer(object):
         for document in self.values():
             yield document
 
-    def add(self, document):
-        """Append a schema document
-
-        :param document: zeep.xsd.schema.SchemaDocument
-
-        """
-        logger.debug("Add document with tns %s to schema %s", document.namespace, id(self))
+    def add(self, document: "SchemaDocument") -> None:
+        """Add a schema document"""
+        logger.debug(
+            "Add document with tns %s to schema %s", document.namespace, id(self)
+        )
         documents = self._instances.setdefault(document.namespace, [])
         documents.append(document)
 
-    def get_all_namespaces(self):
-        return self._instances.keys()
+    def get_all_namespaces(self) -> typing.List[str]:
+        return list(self._instances.keys())
 
-    def get_by_namespace(self, namespace, fail_silently):
+    def get_by_namespace(
+        self, namespace, fail_silently
+    ) -> typing.List["SchemaDocument"]:
         if namespace not in self._instances:
             if fail_silently:
                 return []
             raise exceptions.NamespaceError(
-                "No schema available for the namespace %r" % namespace)
+                "No schema available for the namespace %r" % namespace
+            )
         return self._instances[namespace]
 
-    def get_by_namespace_and_location(self, namespace, location):
-        """Return list of SchemaDocument's for the given namespace AND
-        location.
-
-        :rtype: zeep.xsd.schema.SchemaDocument
-
-        """
+    def get_by_namespace_and_location(
+        self, namespace: str, location: str
+    ) -> typing.Optional["SchemaDocument"]:
+        """Return a SchemaDocument for the given namespace AND location"""
         documents = self.get_by_namespace(namespace, fail_silently=True)
         for document in documents:
             if document._location == location:
                 return document
+        return None
 
-    def has_schema_document_for_ns(self, namespace):
+    def has_schema_document_for_ns(self, namespace: str) -> bool:
         """Return a boolean if there is a SchemaDocument for the namespace.
 
         :rtype: boolean
@@ -372,10 +352,12 @@ class _SchemaContainer(object):
                 yield document
 
 
-class SchemaDocument(object):
+class SchemaDocument:
     """A Schema Document consists of a set of schema components for a
     specific target namespace.
 
+    This represents an xsd:Schema object
+
     """
 
     def __init__(self, namespace, location, base_url):
@@ -396,14 +378,17 @@ class SchemaDocument(object):
         self._types = {}
 
         self._imports = OrderedDict()
-        self._element_form = 'unqualified'
-        self._attribute_form = 'unqualified'
+        self._element_form = "unqualified"
+        self._attribute_form = "unqualified"
         self._resolved = False
         # self._xml_schema = None
 
     def __repr__(self):
-        return '<SchemaDocument(location=%r, tns=%r, is_empty=%r)>' % (
-            self._location, self._target_namespace, self.is_empty)
+        return "<SchemaDocument(location=%r, tns=%r, is_empty=%r)>" % (
+            self._location,
+            self._target_namespace,
+            self.is_empty,
+        )
 
     @property
     def namespace(self):
@@ -425,8 +410,9 @@ class SchemaDocument(object):
 
         if not schema.documents.has_schema_document_for_ns(self._target_namespace):
             raise RuntimeError(
-                "The document needs to be registered in the schema before " +
-                "it can be loaded")
+                "The document needs to be registered in the schema before "
+                + "it can be loaded"
+            )
 
         # Disable XML schema validation for now
         # if len(node) > 0:
@@ -456,12 +442,14 @@ class SchemaDocument(object):
                     (
                         "Unable to resolve %(item_name)s %(qname)s in "
                         "%(file)s. (via %(parent)s)"
-                    ) % {
-                        'item_name': exc.item_name,
-                        'qname': exc.qname,
-                        'file': exc.location,
-                        'parent': obj.qname,
-                    })
+                    )
+                    % {
+                        "item_name": exc.item_name,
+                        "qname": exc.qname,
+                        "file": exc.location,
+                        "parent": obj.qname,
+                    }
+                )
 
         _resolve_dict(self._attribute_groups)
         _resolve_dict(self._attributes)
@@ -482,90 +470,51 @@ class SchemaDocument(object):
     def is_imported(self, namespace):
         return namespace in self._imports
 
-    def register_type(self, qname, value):
-        """Register a xsd.Type in this schema
+    def register_type(self, qname: etree.QName, value: xsd_types.Type):
+        """Register a xsd.Type in this schema"""
+        self._add_component(qname, value, self._types, "type")
 
-        :type qname: str or etree.QName
-        :type value: zeep.xsd.Type
+    def register_element(self, qname: etree.QName, value: xsd_elements.Element):
+        """Register a xsd.Element in this schema"""
+        self._add_component(qname, value, self._elements, "element")
 
-        """
-        self._add_component(qname, value, self._types, 'type')
+    def register_group(self, qname: etree.QName, value: xsd_elements.Group):
+        """Register a xsd:Group in this schema"""
+        self._add_component(qname, value, self._groups, "group")
 
-    def register_element(self, qname, value):
-        """Register a xsd.Element in this schema
+    def register_attribute(self, qname: str, value: xsd_elements.Attribute):
+        """Register a xsd:Attribute in this schema"""
+        self._add_component(qname, value, self._attributes, "attribute")
 
-        :type qname: str or etree.QName
-        :type value: zeep.xsd.Element
+    def register_attribute_group(
+        self, qname: etree.QName, value: xsd_elements.AttributeGroup
+    ) -> None:
+        """Register a xsd:AttributeGroup in this schema"""
+        self._add_component(qname, value, self._attribute_groups, "attribute_group")
 
-        """
-        self._add_component(qname, value, self._elements, 'element')
-
-    def register_group(self, qname, value):
-        """Register a xsd.Element in this schema
-
-        :type qname: str or etree.QName
-        :type value: zeep.xsd.Element
-
-        """
-        self._add_component(qname, value, self._groups, 'group')
-
-    def register_attribute(self, qname, value):
-        """Register a xsd.Element in this schema
-
-        :type qname: str or etree.QName
-        :type value: zeep.xsd.Attribute
-
-        """
-        self._add_component(qname, value, self._attributes, 'attribute')
-
-    def register_attribute_group(self, qname, value):
-        """Register a xsd.Element in this schema
-
-        :type qname: str or etree.QName
-        :type value: zeep.xsd.Group
-
-        """
-        self._add_component(qname, value, self._attribute_groups, 'attribute_group')
-
-    def get_type(self, qname):
+    def get_type(self, qname: etree.QName):
         """Return a xsd.Type object from this schema
 
         :rtype: zeep.xsd.ComplexType or zeep.xsd.AnySimpleType
 
         """
-        return self._get_component(qname, self._types, 'type')
+        return self._get_component(qname, self._types, "type")
 
-    def get_element(self, qname):
-        """Return a xsd.Element object from this schema
+    def get_element(self, qname) -> xsd_elements.Element:
+        """Return a xsd.Element object from this schema"""
+        return self._get_component(qname, self._elements, "element")
 
-        :rtype: zeep.xsd.Element
+    def get_group(self, qname) -> xsd_elements.Group:
+        """Return a xsd.Group object from this schema"""
+        return self._get_component(qname, self._groups, "group")
 
-        """
-        return self._get_component(qname, self._elements, 'element')
+    def get_attribute(self, qname) -> xsd_elements.Attribute:
+        """Return a xsd.Attribute object from this schema"""
+        return self._get_component(qname, self._attributes, "attribute")
 
-    def get_group(self, qname):
-        """Return a xsd.Group object from this schema.
-
-        :rtype: zeep.xsd.Group
-
-        """
-        return self._get_component(qname, self._groups, 'group')
-
-    def get_attribute(self, qname):
-        """Return a xsd.Attribute object from this schema
-
-        :rtype: zeep.xsd.Attribute
-
-        """
-        return self._get_component(qname, self._attributes, 'attribute')
-
-    def get_attribute_group(self, qname):
-        """Return a xsd.AttributeGroup object from this schema
-
-        :rtype: zeep.xsd.AttributeGroup
-
-        """
-        return self._get_component(qname, self._attribute_groups, 'attributeGroup')
+    def get_attribute_group(self, qname) -> xsd_elements.AttributeGroup:
+        """Return a xsd.AttributeGroup object from this schema"""
+        return self._get_component(qname, self._attribute_groups, "attributeGroup")
 
     def _add_component(self, name, value, items, item_name):
         if isinstance(name, etree.QName):
@@ -577,17 +526,20 @@ class SchemaDocument(object):
         try:
             return items[qname]
         except KeyError:
-            known_items = ', '.join(items.keys())
-            raise exceptions.LookupError((
-                "No %(item_name)s '%(localname)s' in namespace %(namespace)s. " +
-                "Available %(item_name_plural)s are: %(known_items)s"
-                ) % {
-                    'item_name': item_name,
-                    'item_name_plural': item_name + 's',
-                    'localname': qname.localname,
-                    'namespace': qname.namespace,
-                    'known_items': known_items or ' - '
+            known_items = ", ".join(items.keys())
+            raise exceptions.LookupError(
+                (
+                    "No %(item_name)s '%(localname)s' in namespace %(namespace)s. "
+                    + "Available %(item_name_plural)s are: %(known_items)s"
+                )
+                % {
+                    "item_name": item_name,
+                    "item_name_plural": item_name + "s",
+                    "localname": qname.localname,
+                    "namespace": qname.namespace,
+                    "known_items": known_items or " - ",
                 },
                 qname=qname,
                 item_name=item_name,
-                location=self._location)
+                location=self._location,
+            )
